type prim = {name : string; arity : int}
let succ = {name ="succ"; arity = 1}
let plus = {name = "plus"; arity = 2}
let minus = {name = "minus"; arity = 2}
let times = {name = "minus"; arity = 2}
let ifz = {name = "ifz"; arity = 3}


type exp =
  | Const of int
  | Prim of prim * exp list
  | Var of var
  | App of exp * exp 
  | Fun of var * exp
  | Funrec of var * var * exp
  | Let of var * exp * exp
and var = string;;
    
type valeur =
  | Vconst of int
  | Vfun of var * exp * env
  | Vfunrec of var * var * exp * env
and env = (var * valeur) list;;
    
type error =
  | Libre of string
  | Delta of prim * valeur list
  | Beta of exp
exception Error of error
let error e = raise (Error e);;

let delta_plus [ Vconst x; Vconst y] =  Vconst (x + y)
let delta_times [ Vconst x; Vconst y] =  Vconst (x * y)
let delta_minus  [ Vconst x; Vconst y] =  Vconst (x - y)
let delta_ifz [ Vconst x; u; v; ] = if x = 0 then u else v
let delta_succ [ Vconst x ] = Vconst (x+1)
let delta =
  [ "succ", delta_succ; "plus", delta_plus;
    "minus", delta_minus; "times", delta_times;
    "ifz", delta_ifz; 
  ] ;;

let find x env =
  try List.assoc x env with Not_found -> error (Libre x);;

let rec eval env = function
  | Const c -> Vconst c
  | Var x -> find x env
  | Fun (x, a) -> Vfun (x, a, env)
  | Funrec (f, x, a) -> Vfunrec (f, x, a, env)
  | App (a1, a2) as e ->
      let f = eval env a1 in
      let v = eval env a2 in
      begin match f with
      | Vfun (x, a, env0) ->
          eval ((x, v) :: env0) a
      | Vfunrec (f, x, a, env0) as r ->
          eval ((f, r) :: (x, v) :: env0) a
      | Vconst _ -> error (Beta e)
      end
  | Prim (f, args) ->
      let vargs = List.map (eval env) args in
      begin try  (List.assoc f.name delta) vargs
      with x -> error (Delta (f, vargs))
      end
  | Let (x, e1, e2) ->
      eval env (App (Fun (x, e2), e1)) ;;


